@@ -22,10 +22,9 @@
 // This file contains a bunch of internal declarations that must appear before rpc.h can start.
 // We don't define these directly in rpc.h because it makes the file hard to read.
 
-#ifndef CAPNP_RPC_PRELUDE_H_
-#define CAPNP_RPC_PRELUDE_H_
+#pragma once
 
-#if defined(__GNUC__) && !CAPNP_HEADER_WARNINGS
+#if defined(__GNUC__) && !defined(CAPNP_HEADER_WARNINGS)
 #pragma GCC system_header
 #endif
 
@@ -43,7 +42,7 @@ class RpcSystem;
 namespace _ {  // private
 
 class VatNetworkBase {
-  // Non-template version of VatNetwork.  Ignore this class; see VatNetwork, below.
+  // Non-template version of VatNetwork.  Ignore this class; see VatNetwork in rpc.h.
 
 public:
   class Connection;
@@ -59,8 +58,9 @@ public:
     virtual kj::Own<OutgoingRpcMessage> newOutgoingMessage(uint firstSegmentWordSize) = 0;
     virtual kj::Promise<kj::Maybe<kj::Own<IncomingRpcMessage>>> receiveIncomingMessage() = 0;
     virtual kj::Promise<void> shutdown() = 0;
+    virtual AnyStruct::Reader baseGetPeerVatId() = 0;
   };
-  virtual kj::Maybe<kj::Own<Connection>> baseConnect(_::StructReader vatId) = 0;
+  virtual kj::Maybe<kj::Own<Connection>> baseConnect(AnyStruct::Reader vatId) = 0;
   virtual kj::Promise<kj::Own<Connection>> baseAccept() = 0;
 };
 
@@ -69,10 +69,20 @@ public:
   virtual Capability::Client baseRestore(AnyPointer::Reader ref) = 0;
 };
 
+class BootstrapFactoryBase {
+  // Non-template version of BootstrapFactory.  Ignore this class; see BootstrapFactory in rpc.h.
+public:
+  virtual Capability::Client baseCreateFor(AnyStruct::Reader clientId) = 0;
+};
+
 class RpcSystemBase {
+  // Non-template version of RpcSystem.  Ignore this class; see RpcSystem in rpc.h.
+
 public:
   RpcSystemBase(VatNetworkBase& network, kj::Maybe<Capability::Client> bootstrapInterface,
                 kj::Maybe<RealmGateway<>::Client> gateway);
+  RpcSystemBase(VatNetworkBase& network, BootstrapFactoryBase& bootstrapFactory,
+                kj::Maybe<RealmGateway<>::Client> gateway);
   RpcSystemBase(VatNetworkBase& network, SturdyRefRestorerBase& restorer);
   RpcSystemBase(RpcSystemBase&& other) noexcept;
   ~RpcSystemBase() noexcept(false);
@@ -81,18 +91,19 @@ private:
   class Impl;
   kj::Own<Impl> impl;
 
-  Capability::Client baseBootstrap(_::StructReader vatId);
-  Capability::Client baseRestore(_::StructReader vatId, AnyPointer::Reader objectId);
-  // TODO(someday):  Maybe define a public API called `TypelessStruct` so we don't have to rely
-  // on `_::StructReader` here?
+  Capability::Client baseBootstrap(AnyStruct::Reader vatId);
+  Capability::Client baseRestore(AnyStruct::Reader vatId, AnyPointer::Reader objectId);
+  void baseSetFlowLimit(size_t words);
 
   template <typename>
   friend class capnp::RpcSystem;
 };
 
 template <typename T> struct InternalRefFromRealmGateway_;
-template <typename InternalRef, typename ExternalRef>
-struct InternalRefFromRealmGateway_<RealmGateway<InternalRef, ExternalRef>> {
+template <typename InternalRef, typename ExternalRef, typename InternalOwner,
+          typename ExternalOwner>
+struct InternalRefFromRealmGateway_<RealmGateway<InternalRef, ExternalRef, InternalOwner,
+                                    ExternalOwner>> {
   typedef InternalRef Type;
 };
 template <typename T>
@@ -101,8 +112,10 @@ template <typename T>
 using InternalRefFromRealmGatewayClient = InternalRefFromRealmGateway<typename T::Calls>;
 
 template <typename T> struct ExternalRefFromRealmGateway_;
-template <typename InternalRef, typename ExternalRef>
-struct ExternalRefFromRealmGateway_<RealmGateway<InternalRef, ExternalRef>> {
+template <typename InternalRef, typename ExternalRef, typename InternalOwner,
+          typename ExternalOwner>
+struct ExternalRefFromRealmGateway_<RealmGateway<InternalRef, ExternalRef, InternalOwner,
+                                    ExternalOwner>> {
   typedef ExternalRef Type;
 };
 template <typename T>
@@ -112,5 +125,3 @@ using ExternalRefFromRealmGatewayClient = ExternalRefFromRealmGateway<typename T
 
 }  // namespace _ (private)
 }  // namespace capnp
-
-#endif  // CAPNP_RPC_PRELUDE_H_
